# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.api_core.retry import Retry

from airflow.providers.google.cloud.operators.managed_kafka import (
    ManagedKafkaCreateClusterOperator,
    ManagedKafkaCreateTopicOperator,
    ManagedKafkaDeleteClusterOperator,
    ManagedKafkaDeleteConsumerGroupOperator,
    ManagedKafkaDeleteTopicOperator,
    ManagedKafkaGetClusterOperator,
    ManagedKafkaGetConsumerGroupOperator,
    ManagedKafkaGetTopicOperator,
    ManagedKafkaListClustersOperator,
    ManagedKafkaListConsumerGroupsOperator,
    ManagedKafkaListTopicsOperator,
    ManagedKafkaUpdateClusterOperator,
    ManagedKafkaUpdateConsumerGroupOperator,
    ManagedKafkaUpdateTopicOperator,
)

MANAGED_KAFKA_PATH = "airflow.providers.google.cloud.operators.managed_kafka.{}"
TIMEOUT = 120
RETRY = mock.MagicMock(Retry)
METADATA = [("key", "value")]

TASK_ID = "test_task_id"
GCP_PROJECT = "test-project"
GCP_LOCATION = "test-location"
GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]

TEST_CLUSTER_ID: str = "test-cluster-id"
TEST_CLUSTER: dict = {
    "gcp_config": {
        "access_config": {
            "network_configs": {
                "subnet": "subnet_value",
            },
        },
    },
    "capacity_config": {
        "vcpu_count": 1094,
        "memory_bytes": 1311,
    },
}
TEST_CLUSTER_UPDATE_MASK: dict = {"paths": ["gcp_config.access_config.network_configs.subnet"]}
TEST_UPDATED_CLUSTER: dict = {
    "gcp_config": {
        "access_config": {
            "network_configs": {
                "subnet": "new_subnet_value",
            },
        },
    },
}

TEST_TOPIC_ID: str = "test-topic-id"
TEST_TOPIC: dict = {
    "partition_count": 1634,
    "replication_factor": 1912,
}
TEST_TOPIC_UPDATE_MASK: dict = {"paths": ["partition_count"]}
TEST_UPDATED_TOPIC: dict = {
    "partition_count": 2000,
    "replication_factor": 1912,
}

TEST_CONSUMER_GROUP_ID: str = "test-consumer-group-id"


class TestManagedKafkaCreateClusterOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Cluster.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaCreateClusterOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster=TEST_CLUSTER,
            cluster_id=TEST_CLUSTER_ID,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.create_cluster.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster=TEST_CLUSTER,
            cluster_id=TEST_CLUSTER_ID,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaListClustersOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ListClustersResponse.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Cluster.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_cluster_dict_mock, to_clusters_dict_mock):
        page_token = "page_token"
        page_size = 42
        filter = "filter"
        order_by = "order_by"

        op = ManagedKafkaListClustersOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            page_size=page_size,
            page_token=page_token,
            filter=filter,
            order_by=order_by,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.list_clusters.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            page_size=page_size,
            page_token=page_token,
            filter=filter,
            order_by=order_by,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaGetClusterOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Cluster.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaGetClusterOperator(
            task_id=TASK_ID,
            cluster_id=TEST_CLUSTER_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.get_cluster.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaUpdateClusterOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Cluster.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaUpdateClusterOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_UPDATED_CLUSTER,
            update_mask=TEST_CLUSTER_UPDATE_MASK,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.update_cluster.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            cluster=TEST_UPDATED_CLUSTER,
            update_mask=TEST_CLUSTER_UPDATE_MASK,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaDeleteClusterOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook):
        op = ManagedKafkaDeleteClusterOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.delete_cluster.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            request_id=None,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaCreateTopicOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Topic.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaCreateTopicOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            topic=TEST_TOPIC,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.create_topic.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            topic=TEST_TOPIC,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaListTopicsOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ListTopicsResponse.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Topic.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_cluster_dict_mock, to_clusters_dict_mock):
        page_token = "page_token"
        page_size = 42

        op = ManagedKafkaListTopicsOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            page_size=page_size,
            page_token=page_token,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.list_topics.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            page_size=page_size,
            page_token=page_token,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaGetTopicOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Topic.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaGetTopicOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.get_topic.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaUpdateTopicOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.Topic.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaUpdateTopicOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            topic=TEST_UPDATED_TOPIC,
            update_mask=TEST_TOPIC_UPDATE_MASK,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.update_topic.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            topic=TEST_UPDATED_TOPIC,
            update_mask=TEST_TOPIC_UPDATE_MASK,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaDeleteTopicOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook):
        op = ManagedKafkaDeleteTopicOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.delete_topic.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            topic_id=TEST_TOPIC_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaListConsumerGroupsOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ListConsumerGroupsResponse.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ConsumerGroup.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_cluster_dict_mock, to_clusters_dict_mock):
        page_token = "page_token"
        page_size = 42

        op = ManagedKafkaListConsumerGroupsOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            page_size=page_size,
            page_token=page_token,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.list_consumer_groups.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            page_size=page_size,
            page_token=page_token,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaGetConsumerGroupOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ConsumerGroup.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaGetConsumerGroupOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.get_consumer_group.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaUpdateConsumerGroupOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("types.ConsumerGroup.to_dict"))
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook, to_dict_mock):
        op = ManagedKafkaUpdateConsumerGroupOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            consumer_group={},
            update_mask={},
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.update_consumer_group.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            consumer_group={},
            update_mask={},
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestManagedKafkaDeleteConsumerGroupOperator:
    @mock.patch(MANAGED_KAFKA_PATH.format("ManagedKafkaHook"))
    def test_execute(self, mock_hook):
        op = ManagedKafkaDeleteConsumerGroupOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        op.execute(context={})
        mock_hook.assert_called_once_with(gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)
        mock_hook.return_value.delete_consumer_group.assert_called_once_with(
            location=GCP_LOCATION,
            project_id=GCP_PROJECT,
            cluster_id=TEST_CLUSTER_ID,
            consumer_group_id=TEST_CONSUMER_GROUP_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Dataflow sensor."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from airflow.providers.common.compat.sdk import AirflowException, BaseSensorOperator, PokeReturnValue, conf
from airflow.providers.google.cloud.hooks.dataflow import (
    DEFAULT_DATAFLOW_LOCATION,
    DataflowHook,
    DataflowJobStatus,
)
from airflow.providers.google.cloud.triggers.dataflow import (
    DataflowJobAutoScalingEventTrigger,
    DataflowJobMessagesTrigger,
    DataflowJobMetricsTrigger,
    DataflowJobStatusTrigger,
)
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class DataflowJobStatusSensor(BaseSensorOperator):
    """
    Checks for the status of a job in Google Cloud Dataflow.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowJobStatusSensor`

    :param job_id: ID of the job to be checked.
    :param expected_statuses: The expected state(s) of the operation.
        See:
        https://cloud.google.com/dataflow/docs/reference/rest/v1b3/projects.jobs#Job.JobState
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: The location of the Dataflow job (for example europe-west1). See:
        https://cloud.google.com/dataflow/docs/concepts/regional-endpoints
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the sensor in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.
    """

    template_fields: Sequence[str] = ("job_id",)

    def __init__(
        self,
        *,
        job_id: str,
        expected_statuses: set[str] | str,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.expected_statuses = (
            {expected_statuses} if isinstance(expected_statuses, str) else expected_statuses
        )
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def poke(self, context: Context) -> bool:
        self.log.info(
            "Waiting for job %s to be in one of the states: %s.",
            self.job_id,
            ", ".join(self.expected_statuses),
        )

        job = self.hook.get_job(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )

        job_status = job["currentState"]
        self.log.debug("Current job status for job %s: %s.", self.job_id, job_status)

        if job_status in self.expected_statuses:
            return True
        if job_status in DataflowJobStatus.TERMINAL_STATES:
            message = f"Job with id '{self.job_id}' is already in terminal state: {job_status}"
            raise AirflowException(message)

        return False

    def execute(self, context: Context) -> None:
        """Airflow runs this method on the worker and defers using the trigger."""
        if not self.deferrable:
            super().execute(context)
        elif not self.poke(context=context):
            self.defer(
                timeout=self.execution_timeout,
                trigger=DataflowJobStatusTrigger(
                    job_id=self.job_id,
                    expected_statuses=self.expected_statuses,
                    project_id=self.project_id,
                    location=self.location,
                    gcp_conn_id=self.gcp_conn_id,
                    poll_sleep=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: dict[str, str | list]) -> bool:
        """
        Execute this method when the task resumes its execution on the worker after deferral.

        Returns True if the trigger returns an event with the success status, otherwise raises
        an exception.
        """
        if event["status"] == "success":
            self.log.info(event["message"])
            return True
        raise AirflowException(f"Sensor failed with the following message: {event['message']}")

    @cached_property
    def hook(self) -> DataflowHook:
        return DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobMetricsSensor(BaseSensorOperator):
    """
    Checks for metrics associated with a single job in Google Cloud Dataflow.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowJobMetricsSensor`

    :param job_id: ID of the job to be checked.
    :param callback: callback which is called with list of read job metrics
        See:
        https://cloud.google.com/dataflow/docs/reference/rest/v1b3/MetricUpdate
    :param fail_on_terminal_state: If set to true sensor will raise Exception when
        job is in terminal state
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: The location of the Dataflow job (for example europe-west1). See:
        https://cloud.google.com/dataflow/docs/concepts/regional-endpoints
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the sensor in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.

    """

    template_fields: Sequence[str] = ("job_id",)

    def __init__(
        self,
        *,
        job_id: str,
        callback: Callable | None = None,
        fail_on_terminal_state: bool = True,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.project_id = project_id
        self.callback = callback
        self.fail_on_terminal_state = fail_on_terminal_state
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def poke(self, context: Context) -> bool:
        if self.fail_on_terminal_state:
            job = self.hook.get_job(
                job_id=self.job_id,
                project_id=self.project_id,
                location=self.location,
            )
            job_status = job["currentState"]
            if job_status in DataflowJobStatus.TERMINAL_STATES:
                message = f"Job with id '{self.job_id}' is already in terminal state: {job_status}"
                raise AirflowException(message)

        result = self.hook.fetch_job_metrics_by_id(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        return result["metrics"] if self.callback is None else self.callback(result["metrics"])

    def execute(self, context: Context) -> Any:
        """Airflow runs this method on the worker and defers using the trigger."""
        if not self.deferrable:
            super().execute(context)
        else:
            self.defer(
                timeout=self.execution_timeout,
                trigger=DataflowJobMetricsTrigger(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                    gcp_conn_id=self.gcp_conn_id,
                    poll_sleep=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                    fail_on_terminal_state=self.fail_on_terminal_state,
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: dict[str, str | list]) -> Any:
        """
        Execute this method when the task resumes its execution on the worker after deferral.

        If the trigger returns an event with success status - passes the event result to the callback function.
        Returns the event result if no callback function is provided.

        If the trigger returns an event with error status - raises an exception.
        """
        if event["status"] == "success":
            self.log.info(event["message"])
            return event["result"] if self.callback is None else self.callback(event["result"])
        raise AirflowException(f"Sensor failed with the following message: {event['message']}")

    @cached_property
    def hook(self) -> DataflowHook:
        return DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobMessagesSensor(BaseSensorOperator):
    """
    Checks for job messages associated with a single job in Google Cloud Dataflow.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowJobMessagesSensor`

    :param job_id: ID of the Dataflow job to be checked.
    :param callback: a function that can accept a list of serialized job messages.
        It can do whatever you want it to do. If the callback function is not provided,
        then on successful completion the task will exit with True value.
        For more info about the job message content see:
        https://cloud.google.com/python/docs/reference/dataflow/latest/google.cloud.dataflow_v1beta3.types.JobMessage
    :param fail_on_terminal_state: If set to True the sensor will raise an exception when the job reaches a terminal state.
        No job messages will be returned.
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: The location of the Dataflow job (for example europe-west1).
        If set to None then the value of DEFAULT_DATAFLOW_LOCATION will be used.
        See: https://cloud.google.com/dataflow/docs/concepts/regional-endpoints
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the sensor in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.
    """

    template_fields: Sequence[str] = ("job_id",)

    def __init__(
        self,
        *,
        job_id: str,
        callback: Callable | None = None,
        fail_on_terminal_state: bool = True,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.project_id = project_id
        self.callback = callback
        self.fail_on_terminal_state = fail_on_terminal_state
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def poke(self, context: Context) -> PokeReturnValue | bool:
        if self.fail_on_terminal_state:
            job = self.hook.get_job(
                job_id=self.job_id,
                project_id=self.project_id,
                location=self.location,
            )
            job_status = job["currentState"]
            if job_status in DataflowJobStatus.TERMINAL_STATES:
                message = f"Job with id '{self.job_id}' is already in terminal state: {job_status}"
                raise AirflowException(message)

        result = self.hook.fetch_job_messages_by_id(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        result = result if self.callback is None else self.callback(result)

        if isinstance(result, PokeReturnValue):
            return result

        if bool(result):
            return PokeReturnValue(
                is_done=True,
                xcom_value=result,
            )
        return False

    def execute(self, context: Context) -> Any:
        """Airflow runs this method on the worker and defers using the trigger."""
        if not self.deferrable:
            super().execute(context)
        else:
            self.defer(
                timeout=self.execution_timeout,
                trigger=DataflowJobMessagesTrigger(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                    gcp_conn_id=self.gcp_conn_id,
                    poll_sleep=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                    fail_on_terminal_state=self.fail_on_terminal_state,
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: dict[str, str | list]) -> Any:
        """
        Execute this method when the task resumes its execution on the worker after deferral.

        If the trigger returns an event with success status - passes the event result to the callback function.
        Returns the event result if no callback function is provided.

        If the trigger returns an event with error status - raises an exception.
        """
        if event["status"] == "success":
            self.log.info(event["message"])
            return event["result"] if self.callback is None else self.callback(event["result"])
        raise AirflowException(f"Sensor failed with the following message: {event['message']}")

    @cached_property
    def hook(self) -> DataflowHook:
        return DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )


class DataflowJobAutoScalingEventsSensor(BaseSensorOperator):
    """
    Checks for autoscaling events associated with a single job in Google Cloud Dataflow.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowJobAutoScalingEventsSensor`

    :param job_id: ID of the Dataflow job to be checked.
    :param callback: a function that can accept a list of serialized autoscaling events.
        It can do whatever you want it to do. If the callback function is not provided,
        then on successful completion the task will exit with True value.
        For more info about the autoscaling event content see:
        https://cloud.google.com/python/docs/reference/dataflow/latest/google.cloud.dataflow_v1beta3.types.AutoscalingEvent
    :param fail_on_terminal_state: If set to True the sensor will raise an exception when the job reaches a terminal state.
        No autoscaling events will be returned.
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: The location of the Dataflow job (for example europe-west1).
        If set to None then the value of DEFAULT_DATAFLOW_LOCATION will be used.
        See: https://cloud.google.com/dataflow/docs/concepts/regional-endpoints
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: If True, run the sensor in the deferrable mode.
    :param poll_interval: Time (seconds) to wait between two consecutive calls to check the job.
    """

    template_fields: Sequence[str] = ("job_id",)

    def __init__(
        self,
        *,
        job_id: str,
        callback: Callable | None = None,
        fail_on_terminal_state: bool = True,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 60,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job_id = job_id
        self.project_id = project_id
        self.callback = callback
        self.fail_on_terminal_state = fail_on_terminal_state
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

    def poke(self, context: Context) -> PokeReturnValue | bool:
        if self.fail_on_terminal_state:
            job = self.hook.get_job(
                job_id=self.job_id,
                project_id=self.project_id,
                location=self.location,
            )
            job_status = job["currentState"]
            if job_status in DataflowJobStatus.TERMINAL_STATES:
                message = f"Job with id '{self.job_id}' is already in terminal state: {job_status}"
                raise AirflowException(message)

        result = self.hook.fetch_job_autoscaling_events_by_id(
            job_id=self.job_id,
            project_id=self.project_id,
            location=self.location,
        )
        result = result if self.callback is None else self.callback(result)
        if isinstance(result, PokeReturnValue):
            return result

        if bool(result):
            return PokeReturnValue(
                is_done=True,
                xcom_value=result,
            )
        return False

    def execute(self, context: Context) -> Any:
        """Airflow runs this method on the worker and defers using the trigger."""
        if not self.deferrable:
            super().execute(context)
        else:
            self.defer(
                trigger=DataflowJobAutoScalingEventTrigger(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.location,
                    gcp_conn_id=self.gcp_conn_id,
                    poll_sleep=self.poll_interval,
                    impersonation_chain=self.impersonation_chain,
                    fail_on_terminal_state=self.fail_on_terminal_state,
                ),
                method_name="execute_complete",
            )

    def execute_complete(self, context: Context, event: dict[str, str | list]) -> Any:
        """
        Execute this method when the task resumes its execution on the worker after deferral.

        If the trigger returns an event with success status - passes the event result to the callback function.
        Returns the event result if no callback function is provided.

        If the trigger returns an event with error status - raises an exception.
        """
        if event["status"] == "success":
            self.log.info(event["message"])
            return event["result"] if self.callback is None else self.callback(event["result"])
        raise AirflowException(f"Sensor failed with the following message: {event['message']}")

    @cached_property
    def hook(self) -> DataflowHook:
        return DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google DisplayVideo operators."""

from __future__ import annotations

import os
import tempfile
import zipfile
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.marketing_platform.hooks.display_video import GoogleDisplayVideo360Hook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class GoogleDisplayVideo360CreateSDFDownloadTaskOperator(BaseOperator):
    """
    Creates an SDF operation task.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360CreateSDFDownloadTaskOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/display-video/api/reference/rest`

    :param version: The SDF version of the downloaded file.
    :param partner_id: The ID of the partner to download SDF for.
    :param advertiser_id: The ID of the advertiser to download SDF for.
    :param parent_entity_filter: Filters on selected file types.
    :param id_filter: Filters on entities by their entity IDs.
    :param inventory_source_filter: Filters on Inventory Sources by their IDs.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body_request",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        body_request: dict[str, Any],
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.body_request = body_request
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict[str, Any]:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Creating operation for SDF download task...")
        operation = hook.create_sdf_download_operation(body_request=self.body_request)

        name = operation["name"]
        context["task_instance"].xcom_push(key="name", value=name)
        self.log.info("Created SDF operation with name: %s", name)

        return operation


class GoogleDisplayVideo360SDFtoGCSOperator(BaseOperator):
    """
    Download SDF media and save it in the Google Cloud Storage.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360SDFtoGCSOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/display-video/api/reference/rest`

    :param version: The SDF version of the downloaded file.
    :param partner_id: The ID of the partner to download SDF for.
    :param advertiser_id: The ID of the advertiser to download SDF for.
    :param parent_entity_filter: Filters on selected file types.
    :param id_filter: Filters on entities by their entity IDs.
    :param inventory_source_filter: Filters on Inventory Sources by their IDs.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "operation_name",
        "bucket_name",
        "object_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        operation_name: str,
        bucket_name: str,
        object_name: str,
        gzip: bool = False,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.operation_name = operation_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.gzip = gzip
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> str:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Retrieving operation...")
        operation_state = hook.get_sdf_download_operation(operation_name=self.operation_name)

        self.log.info("Creating file for upload...")
        media = hook.download_media(resource_name=operation_state["response"]["resourceName"])

        self.log.info("Sending file to the Google Cloud Storage...")
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_path = os.path.join(tmp_dir, "sdf.zip")

            # Download the ZIP
            with open(zip_path, "wb") as f:
                hook.download_content_from_request(f, media, chunk_size=1024 * 1024)

            # Extract CSV
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(tmp_dir)

            # Upload CSV file
            for fname in os.listdir(tmp_dir):
                if fname.endswith(".csv"):
                    gcs_hook.upload(
                        bucket_name=self.bucket_name,
                        object_name=self.object_name,
                        filename=os.path.join(tmp_dir, fname),
                        gzip=False,
                    )
        return f"{self.bucket_name}/{self.object_name}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.google.cloud.utils.field_validator import (
    GcpBodyFieldValidator,
    GcpFieldValidationException,
    GcpValidationSpecificationException,
)


class TestGcpBodyFieldValidator:
    def test_validate_should_not_raise_exception_if_field_and_body_are_both_empty(self):
        specification = []
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_fail_if_body_is_none(self):
        specification = []
        body = None

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(
            RuntimeError, match="The body to validate is `None`. Please provide a dictionary to validate."
        ):
            validator.validate(body)

    def test_validate_should_fail_if_specification_is_none(self):
        specification = None
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(TypeError):
            validator.validate(body)

    def test_validate_should_raise_exception_name_attribute_is_missing_from_specs(self):
        specification = [dict(allow_empty=False)]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(KeyError):
            validator.validate(body)

    def test_validate_should_raise_exception_if_field_is_not_present(self):
        specification = [dict(name="name", allow_empty=False)]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_validate_a_single_field(self):
        specification = [dict(name="name", allow_empty=False)]
        body = {"name": "bigquery"}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_fail_if_body_is_not_a_dict(self):
        specification = [dict(name="name", allow_empty=False)]
        body = [{"name": "bigquery"}]

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(AttributeError):
            validator.validate(body)

    def test_validate_should_fail_for_set_allow_empty_when_field_is_none(self):
        specification = [dict(name="name", allow_empty=True)]
        body = {"name": None}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_interpret_allow_empty_clause(self):
        specification = [dict(name="name", allow_empty=True)]
        body = {"name": ""}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_raise_if_empty_clause_is_false(self):
        specification = [dict(name="name", allow_empty=False)]
        body = {"name": None}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_raise_if_version_mismatch_is_found(self):
        specification = [dict(name="name", allow_empty=False, api_version="v2")]
        body = {"name": "value"}

        validator = GcpBodyFieldValidator(specification, "v1")

        validator.validate(body)

    def test_validate_should_interpret_optional_irrespective_of_allow_empty(self):
        specification = [dict(name="name", allow_empty=False, optional=True)]
        body = {"name": None}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_interpret_optional_clause(self):
        specification = [dict(name="name", allow_empty=False, optional=True)]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_raise_exception_if_optional_clause_is_false_and_field_not_present(self):
        specification = [dict(name="name", allow_empty=False, optional=False)]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_interpret_dict_type(self):
        specification = [dict(name="labels", optional=True, type="dict")]
        body = {"labels": {"one": "value"}}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_fail_if_value_is_not_dict_as_per_specs(self):
        specification = [dict(name="labels", optional=True, type="dict")]
        body = {"labels": 1}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_not_allow_both_type_and_allow_empty_in_a_spec(self):
        specification = [dict(name="labels", optional=True, type="dict", allow_empty=True)]
        body = {"labels": 1}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpValidationSpecificationException):
            validator.validate(body)

    def test_validate_should_allow_type_and_optional_in_a_spec(self):
        specification = [dict(name="labels", optional=True, type="dict")]
        body = {"labels": {}}

        validator = GcpBodyFieldValidator(specification, "v1")

        assert validator.validate(body) is None

    def test_validate_should_fail_if_union_field_is_not_found(self):
        specification = [
            dict(
                name="an_union",
                type="union",
                optional=False,
                fields=[
                    dict(name="variant_1", regexp=r"^.+$", optional=False, allow_empty=False),
                ],
            )
        ]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")
        assert validator.validate(body) is None

    def test_validate_should_fail_if_there_is_no_nested_field_for_union(self):
        specification = [dict(name="an_union", type="union", optional=False, fields=[])]
        body = {}

        validator = GcpBodyFieldValidator(specification, "v1")

        with pytest.raises(GcpValidationSpecificationException):
            validator.validate(body)

    def test_validate_should_interpret_union_with_one_field(self):
        specification = [
            dict(
                name="an_union",
                type="union",
                fields=[
                    dict(name="variant_1", regexp=r"^.+$"),
                ],
            )
        ]
        body = {"variant_1": "abc", "variant_2": "def"}

        validator = GcpBodyFieldValidator(specification, "v1")
        assert validator.validate(body) is None

    def test_validate_should_fail_if_both_field_of_union_is_present(self):
        specification = [
            dict(
                name="an_union",
                type="union",
                fields=[
                    dict(name="variant_1", regexp=r"^.+$"),
                    dict(name="variant_2", regexp=r"^.+$"),
                ],
            )
        ]
        body = {"variant_1": "abc", "variant_2": "def"}

        validator = GcpBodyFieldValidator(specification, "v1")
        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_validate_when_value_matches_regex(self):
        specification = [
            dict(
                name="an_union",
                type="union",
                fields=[
                    dict(name="variant_1", regexp=r"[^a-z]"),
                ],
            )
        ]
        body = {"variant_1": "12"}

        validator = GcpBodyFieldValidator(specification, "v1")
        assert validator.validate(body) is None

    def test_validate_should_fail_when_value_does_not_match_regex(self):
        specification = [
            dict(
                name="an_union",
                type="union",
                fields=[
                    dict(name="variant_1", regexp=r"[^a-z]"),
                ],
            )
        ]
        body = {"variant_1": "abc"}

        validator = GcpBodyFieldValidator(specification, "v1")
        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_raise_if_custom_validation_is_not_true(self):
        def _int_equal_to_zero(value):
            if int(value) != 0:
                raise GcpFieldValidationException("The available memory has to be equal to 0")

        specification = [dict(name="availableMemoryMb", custom_validation=_int_equal_to_zero)]
        body = {"availableMemoryMb": 1}

        validator = GcpBodyFieldValidator(specification, "v1")
        with pytest.raises(GcpFieldValidationException):
            validator.validate(body)

    def test_validate_should_not_raise_if_custom_validation_is_true(self):
        def _int_equal_to_zero(value):
            if int(value) != 0:
                raise GcpFieldValidationException("The available memory has to be equal to 0")

        specification = [dict(name="availableMemoryMb", custom_validation=_int_equal_to_zero)]
        body = {"availableMemoryMb": 0}

        validator = GcpBodyFieldValidator(specification, "v1")
        assert validator.validate(body) is None

    def test_validate_should_validate_group_of_specs(self):
        specification = [
            dict(name="name", allow_empty=False),
            dict(name="description", allow_empty=False, optional=True),
            dict(name="labels", optional=True, type="dict"),
            dict(
                name="an_union",
                type="union",
                fields=[
                    dict(name="variant_1", regexp=r"^.+$"),
                    dict(name="variant_2", regexp=r"^.+$", api_version="v1beta2"),
                    dict(name="variant_3", type="dict", fields=[dict(name="url", regexp=r"^.+$")]),
                    dict(name="variant_4"),
                ],
            ),
        ]
        body = {"variant_1": "abc", "name": "bigquery"}

        validator = GcpBodyFieldValidator(specification, "v1")
        validator.validate(body)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

from airflow.providers.google.cloud.transfers.bigquery_to_mssql import BigQueryToMsSqlOperator

TASK_ID = "test-bq-create-table-operator"
TEST_DATASET = "test-dataset"
TEST_TABLE_ID = "test-table-id"
TEST_DAG_ID = "test-bigquery-operators"
TEST_PROJECT = "test-project"


def _make_bq_table(schema_names: list[str]):
    class TableObj:
        def __init__(self, schema):
            self.schema = []
            for n in schema:
                field = MagicMock()
                field.name = n
                self.schema.append(field)
            self.description = "table description"
            self.external_data_configuration = None
            self.labels = {}
            self.num_rows = 0
            self.num_bytes = 0
            self.table_type = "TABLE"

    return TableObj(schema_names)


class TestBigQueryToMsSqlOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mssql.BigQueryTableLink")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_execute_good_request_to_bq(self, mock_hook, mock_link):
        destination_table = "table"
        operator = BigQueryToMsSqlOperator(
            task_id=TASK_ID,
            source_project_dataset_table=f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=destination_table,
            replace=False,
        )

        operator.execute(None)
        mock_hook.return_value.list_rows.assert_called_once_with(
            dataset_id=TEST_DATASET,
            table_id=TEST_TABLE_ID,
            max_results=1000,
            selected_fields=None,
            start_index=0,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mssql.MsSqlHook")
    def test_get_sql_hook(self, mock_hook):
        hook_expected = mock_hook.return_value

        destination_table = "table"
        operator = BigQueryToMsSqlOperator(
            task_id=TASK_ID,
            source_project_dataset_table=f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=destination_table,
            replace=False,
        )

        hook_actual = operator.get_sql_hook()

        assert hook_actual == hook_expected
        mock_hook.assert_called_once_with(schema=operator.database, mssql_conn_id=operator.mssql_conn_id)

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mssql.BigQueryTableLink")
    def test_persist_links(self, mock_link):
        mock_context = mock.MagicMock()

        destination_table = "table"
        operator = BigQueryToMsSqlOperator(
            task_id=TASK_ID,
            source_project_dataset_table=f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=destination_table,
            replace=False,
        )
        operator.persist_links(context=mock_context)

        mock_link.persist.assert_called_once_with(
            context=mock_context,
            dataset_id=TEST_DATASET,
            project_id=TEST_PROJECT,
            table_id=TEST_TABLE_ID,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mssql.MsSqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_get_openlineage_facets_on_complete_no_selected_fields(self, mock_bq_hook, mock_mssql_hook):
        mock_bq_client = MagicMock()
        table_obj = _make_bq_table(["id", "name", "value"])
        mock_bq_client.get_table.return_value = table_obj
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="mssql", authority="localhost:1433", database="mydb")
        mock_mssql_hook.get_openlineage_database_info.return_value = db_info
        mock_mssql_hook.get_openlineage_default_schema.return_value = "dbo"
        mock_mssql_hook.return_value = mock_mssql_hook

        op = BigQueryToMsSqlOperator(
            task_id="test",
            source_project_dataset_table="proj.dataset.table",
            target_table_name="dbo.destination",
            selected_fields=None,
            database="mydb",
        )
        op.bigquery_hook = mock_bq_hook
        op.mssql_hook = mock_mssql_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(task_instance=MagicMock())
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert input_ds.name == "proj.dataset.table"

        assert "schema" in input_ds.facets
        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name", "value"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "mssql://localhost:1433"
        assert output_ds.name == "mydb.dbo.destination"

        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name", "value"}

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mssql.MsSqlHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_get_openlineage_facets_on_complete_selected_fields(self, mock_bq_hook, mock_mssql_hook):
        mock_bq_client = MagicMock()
        table_obj = _make_bq_table(["id", "name", "value"])
        mock_bq_client.get_table.return_value = table_obj
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="mssql", authority="server.example:1433", database="mydb")
        mock_mssql_hook.get_openlineage_database_info.return_value = db_info
        mock_mssql_hook.get_openlineage_default_schema.return_value = "dbo"
        mock_mssql_hook.return_value = mock_mssql_hook

        op = BigQueryToMsSqlOperator(
            task_id="test",
            source_project_dataset_table="proj.dataset.table",
            target_table_name="dbo.destination",
            selected_fields=["id", "name"],
            database="mydb",
        )
        op.bigquery_hook = mock_bq_hook
        op.mssql_hook = mock_mssql_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(task_instance=MagicMock())
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert "schema" in input_ds.facets

        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "mssql://server.example:1433"
        assert output_ds.name == "mydb.dbo.destination"

        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name"}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google BigQuery Data Transfer Service operators."""

from __future__ import annotations

import time
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.bigquery_datatransfer_v1 import (
    StartManualTransferRunsResponse,
    TransferConfig,
    TransferRun,
    TransferState,
)

from airflow.providers.common.compat.sdk import AirflowException, conf
from airflow.providers.google.cloud.hooks.bigquery_dts import BiqQueryDataTransferServiceHook, get_object_id
from airflow.providers.google.cloud.links.bigquery_dts import BigQueryDataTransferConfigLink
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.bigquery_dts import BigQueryDataTransferRunTrigger
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from google.api_core.retry import Retry

    from airflow.providers.common.compat.sdk import Context


def _get_transfer_config_details(config_transfer_name: str):
    config_details = config_transfer_name.split("/")
    return {"project_id": config_details[1], "region": config_details[3], "config_id": config_details[5]}


class BigQueryCreateDataTransferOperator(GoogleCloudBaseOperator):
    """
    Creates a new data transfer configuration.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:BigQueryCreateDataTransferOperator`

    :param transfer_config: Data transfer configuration to create.
    :param project_id: The BigQuery project id where the transfer configuration should be
            created. If set to None or missing, the default project_id from the Google Cloud connection
            is used.
    :param location: BigQuery Transfer Service location for regional transfers.
    :param authorization_code: authorization code to use with this transfer configuration.
        This is required if new credentials are needed.
    :param retry: A retry object used to retry requests. If `None` is
        specified, requests will not be retried.
    :param timeout: The amount of time, in seconds, to wait for the request to
        complete. Note that if retry is specified, the timeout applies to each individual
        attempt.
    :param metadata: Additional metadata that is provided to the method.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "transfer_config",
        "project_id",
        "authorization_code",
        "gcp_conn_id",
        "impersonation_chain",
    )
    operator_extra_links = (BigQueryDataTransferConfigLink(),)

    def __init__(
        self,
        *,
        transfer_config: dict,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str | None = None,
        authorization_code: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id="google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.transfer_config = transfer_config
        self.authorization_code = authorization_code
        self.project_id = project_id
        self.location = location
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = BiqQueryDataTransferServiceHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain, location=self.location
        )
        self.log.info("Creating DTS transfer config")
        response = hook.create_transfer_config(
            project_id=self.project_id,
            transfer_config=self.transfer_config,
            authorization_code=self.authorization_code,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        transfer_config = _get_transfer_config_details(response.name)
        BigQueryDataTransferConfigLink.persist(
            context=context,
            region=transfer_config["region"],
            config_id=transfer_config["config_id"],
            project_id=transfer_config["project_id"],
        )

        result = TransferConfig.to_dict(response)
        self.log.info("Created DTS transfer config %s", get_object_id(result))
        context["ti"].xcom_push(key="transfer_config_id", value=get_object_id(result))
        # don't push AWS secret in XCOM
        result.get("params", {}).pop("secret_access_key", None)
        result.get("params", {}).pop("access_key_id", None)
        return result


class BigQueryDeleteDataTransferConfigOperator(GoogleCloudBaseOperator):
    """
    Deletes transfer configuration.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:BigQueryDeleteDataTransferConfigOperator`

    :param transfer_config_id: Id of transfer config to be used.
    :param project_id: The BigQuery project id where the transfer configuration should be
        created. If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: BigQuery Transfer Service location for regional transfers.
    :param retry: A retry object used to retry requests. If `None` is
        specified, requests will not be retried.
    :param timeout: The amount of time, in seconds, to wait for the request to
        complete. Note that if retry is specified, the timeout applies to each individual
        attempt.
    :param metadata: Additional metadata that is provided to the method.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "transfer_config_id",
        "project_id",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        transfer_config_id: str,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id="google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.transfer_config_id = transfer_config_id
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = BiqQueryDataTransferServiceHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain, location=self.location
        )
        hook.delete_transfer_config(
            transfer_config_id=self.transfer_config_id,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )


class BigQueryDataTransferServiceStartTransferRunsOperator(GoogleCloudBaseOperator):
    """
    Start manual transfer runs to be executed now with schedule_time equal to current time.

    The transfer runs can be created for a time range where the run_time is between
    start_time (inclusive) and end_time (exclusive), or for a specific run_time.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:BigQueryDataTransferServiceStartTransferRunsOperator`

    :param transfer_config_id: Id of transfer config to be used.
    :param requested_time_range: Time range for the transfer runs that should be started.
        If a dict is provided, it must be of the same form as the protobuf
        message `~google.cloud.bigquery_datatransfer_v1.types.TimeRange`
    :param requested_run_time: Specific run_time for a transfer run to be started. The
        requested_run_time must not be in the future.  If a dict is provided, it
        must be of the same form as the protobuf message
        `~google.cloud.bigquery_datatransfer_v1.types.Timestamp`
    :param project_id: The BigQuery project id where the transfer configuration should be
        created.
    :param location: BigQuery Transfer Service location for regional transfers.
    :param retry: A retry object used to retry requests. If `None` is
        specified, requests will not be retried.
    :param timeout: The amount of time, in seconds, to wait for the request to
        complete. Note that if retry is specified, the timeout applies to each individual
        attempt.
    :param metadata: Additional metadata that is provided to the method.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: Run operator in the deferrable mode.
    """

    template_fields: Sequence[str] = (
        "transfer_config_id",
        "project_id",
        "requested_time_range",
        "requested_run_time",
        "gcp_conn_id",
        "impersonation_chain",
    )
    operator_extra_links = (BigQueryDataTransferConfigLink(),)

    def __init__(
        self,
        *,
        transfer_config_id: str,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str | None = None,
        requested_time_range: dict | None = None,
        requested_run_time: dict | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id="google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.transfer_config_id = transfer_config_id
        self.requested_time_range = requested_time_range
        self.requested_run_time = requested_run_time
        self.retry = retry
        self.timeout = timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self._transfer_run: dict = {}

    @cached_property
    def hook(self) -> BiqQueryDataTransferServiceHook:
        hook = BiqQueryDataTransferServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
            location=self.location,
        )
        return hook

    def execute(self, context: Context):
        self.log.info("Submitting manual transfer for %s", self.transfer_config_id)

        if self.requested_run_time and isinstance(self.requested_run_time.get("seconds"), str):
            self.requested_run_time["seconds"] = int(self.requested_run_time["seconds"])

        response = self.hook.start_manual_transfer_runs(
            transfer_config_id=self.transfer_config_id,
            requested_time_range=self.requested_time_range,
            requested_run_time=self.requested_run_time,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.timeout,
            metadata=self.metadata,
        )

        transfer_config = _get_transfer_config_details(response.runs[0].name)
        BigQueryDataTransferConfigLink.persist(
            context=context,
            region=transfer_config["region"],
            config_id=transfer_config["config_id"],
            project_id=transfer_config["project_id"],
        )

        result = StartManualTransferRunsResponse.to_dict(response)
        run_id = get_object_id(result["runs"][0])
        context["ti"].xcom_push(key="run_id", value=run_id)

        if not self.deferrable:
            # Save as attribute for further use by OpenLineage
            self._transfer_run = self._wait_for_transfer_to_be_done(
                run_id=run_id,
                transfer_config_id=transfer_config["config_id"],
            )
            self.log.info("Transfer run %s submitted successfully.", run_id)
            return self._transfer_run

        self.defer(
            trigger=BigQueryDataTransferRunTrigger(
                project_id=self.project_id,
                config_id=transfer_config["config_id"],
                run_id=run_id,
                gcp_conn_id=self.gcp_conn_id,
                location=self.location,
                impersonation_chain=self.impersonation_chain,
            ),
            method_name="execute_completed",
        )

    def _wait_for_transfer_to_be_done(self, run_id: str, transfer_config_id: str, interval: int = 10):
        if interval <= 0:
            raise ValueError("Interval must be > 0")

        while True:
            transfer_run: TransferRun = self.hook.get_transfer_run(
                run_id=run_id,
                transfer_config_id=transfer_config_id,
                project_id=self.project_id,
                retry=self.retry,
                timeout=self.timeout,
                metadata=self.metadata,
            )
            state = transfer_run.state

            if self._job_is_done(state):
                if state in (TransferState.FAILED, TransferState.CANCELLED):
                    raise AirflowException(f"Transfer run was finished with {state} status.")

                result = TransferRun.to_dict(transfer_run)
                return result

            self.log.info("Transfer run is still working, waiting for %s seconds...", interval)
            self.log.info("Transfer run status: %s", state)
            time.sleep(interval)

    @staticmethod
    def _job_is_done(state: TransferState) -> bool:
        finished_job_statuses = [
            state.SUCCEEDED,
            state.CANCELLED,
            state.FAILED,
        ]

        return state in finished_job_statuses

    def execute_completed(self, context: Context, event: dict):
        """Execute after invoked trigger in defer method finishes its job."""
        if event["status"] in ("failed", "cancelled"):
            self.log.error("Trigger finished its work with status: %s.", event["status"])
            raise AirflowException(event["message"])

        transfer_run: TransferRun = self.hook.get_transfer_run(
            project_id=self.project_id,
            run_id=event["run_id"],
            transfer_config_id=event["config_id"],
        )

        self.log.info(
            "%s finished with message: %s",
            event["run_id"],
            event["message"],
        )

        # Save as attribute for further use by OpenLineage
        self._transfer_run = TransferRun.to_dict(transfer_run)
        return self._transfer_run

    def get_openlineage_facets_on_complete(self, _):
        """Implement _on_complete as we need a run config to extract information."""
        from urllib.parse import urlsplit

        from airflow.providers.common.compat.openlineage.facet import Dataset, ErrorMessageRunFacet
        from airflow.providers.google.cloud.hooks.gcs import _parse_gcs_url
        from airflow.providers.google.cloud.openlineage.utils import (
            BIGQUERY_NAMESPACE,
            extract_ds_name_from_gcs_path,
        )
        from airflow.providers.openlineage.extractors import OperatorLineage
        from airflow.providers.openlineage.sqlparser import DatabaseInfo, SQLParser

        if not self._transfer_run:
            self.log.debug("No BigQuery Data Transfer configuration was found by OpenLineage.")
            return OperatorLineage()

        data_source_id = self._transfer_run["data_source_id"]
        dest_dataset_id = self._transfer_run["destination_dataset_id"]
        params = self._transfer_run["params"]

        input_datasets, output_datasets = [], []
        run_facets, job_facets = {}, {}
        if data_source_id in ("google_cloud_storage", "amazon_s3", "azure_blob_storage"):
            if data_source_id == "google_cloud_storage":
                bucket, path = _parse_gcs_url(params["data_path_template"])  # gs://bucket...
                namespace = f"gs://{bucket}"
                name = extract_ds_name_from_gcs_path(path)
            elif data_source_id == "amazon_s3":
                parsed_url = urlsplit(params["data_path"])  # s3://bucket...
                namespace = f"s3://{parsed_url.netloc}"
                name = extract_ds_name_from_gcs_path(parsed_url.path)
            else:  # azure_blob_storage
                storage_account = params["storage_account"]
                container = params["container"]
                namespace = f"abfss://{container}@{storage_account}.dfs.core.windows.net"
                name = extract_ds_name_from_gcs_path(params["data_path"])

            input_datasets.append(Dataset(namespace=namespace, name=name))
            dest_table_name = params["destination_table_name_template"]
            output_datasets.append(
                Dataset(
                    namespace=BIGQUERY_NAMESPACE,
                    name=f"{self.project_id}.{dest_dataset_id}.{dest_table_name}",
                )
            )
        elif data_source_id in ("postgresql", "oracle", "mysql"):
            scheme = data_source_id if data_source_id != "postgresql" else "postgres"
            host = params["connector.endpoint.host"]
            port = params["connector.endpoint.port"]

            for asset in params["assets"]:
                # MySQL: db/table; Other: db/schema/table;
                table_name = asset.split("/")[-1]

                input_datasets.append(
                    Dataset(namespace=f"{scheme}://{host}:{int(port)}", name=asset.replace("/", "."))
                )
                output_datasets.append(
                    Dataset(
                        namespace=BIGQUERY_NAMESPACE, name=f"{self.project_id}.{dest_dataset_id}.{table_name}"
                    )
                )
        elif data_source_id == "scheduled_query":
            bq_db_info = DatabaseInfo(
                scheme="bigquery",
                authority=None,
                database=self.project_id,
            )
            parser_result = SQLParser("bigquery").generate_openlineage_metadata_from_sql(
                sql=params["query"],
                database_info=bq_db_info,
                database=self.project_id,
                use_connection=False,
                hook=None,  # Hook is not used when use_connection=False
                sqlalchemy_engine=None,
            )
            if parser_result.inputs:
                input_datasets.extend(parser_result.inputs)
            if parser_result.outputs:
                output_datasets.extend(parser_result.outputs)
            if parser_result.job_facets:
                job_facets = {**job_facets, **parser_result.job_facets}
            if parser_result.run_facets:
                run_facets = {**run_facets, **parser_result.run_facets}
            dest_table_name = params.get("destination_table_name_template")
            if dest_table_name:
                output_datasets.append(
                    Dataset(
                        namespace=BIGQUERY_NAMESPACE,
                        name=f"{self.project_id}.{dest_dataset_id}.{dest_table_name}",
                    )
                )
        else:
            self.log.debug(
                "BigQuery Data Transfer data_source_id `%s` is not supported by OpenLineage.", data_source_id
            )
            return OperatorLineage()

        error_status = self._transfer_run.get("error_status")
        if error_status and str(error_status["code"]) != "0":
            run_facets["errorMessage"] = ErrorMessageRunFacet(
                message=error_status["message"],
                programmingLanguage="python",
                stackTrace=str(error_status["details"]),
            )

        return OperatorLineage(
            inputs=input_datasets, outputs=output_datasets, job_facets=job_facets, run_facets=run_facets
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.workflows.executions_v1beta import Execution

from airflow.providers.common.compat.sdk import AirflowException, BaseSensorOperator
from airflow.providers.google.cloud.hooks.workflows import WorkflowsHook
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from google.api_core.retry import Retry

    from airflow.providers.common.compat.sdk import Context


class WorkflowExecutionSensor(BaseSensorOperator):
    """
    Checks state of an execution for the given ``workflow_id`` and ``execution_id``.

    :param workflow_id: Required. The ID of the workflow.
    :param execution_id: Required. The ID of the execution.
    :param project_id: Required. The ID of the Google Cloud project the cluster belongs to.
    :param location: Required. The Cloud Dataproc region in which to handle the request.
    :param success_states: Execution states to be considered as successful, by default
        it's only ``SUCCEEDED`` state
    :param failure_states: Execution states to be considered as failures, by default
        they are ``FAILED`` and ``CANCELLED`` states.
    :param retry: A retry object used to retry requests. If ``None`` is specified, requests will not be
        retried.
    :param request_timeout: The amount of time, in seconds, to wait for the request to complete. Note that if
        ``retry`` is specified, the timeout applies to each individual attempt.
    :param metadata: Additional metadata that is provided to the method.
    """

    template_fields: Sequence[str] = ("location", "workflow_id", "execution_id")

    def __init__(
        self,
        *,
        workflow_id: str,
        execution_id: str,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        success_states: set[Execution.State] | None = None,
        failure_states: set[Execution.State] | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        request_timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.success_states = success_states or {Execution.State(Execution.State.SUCCEEDED)}
        self.failure_states = failure_states or {
            Execution.State(Execution.State.FAILED),
            Execution.State(Execution.State.CANCELLED),
        }
        self.workflow_id = workflow_id
        self.execution_id = execution_id
        self.location = location
        self.project_id = project_id
        self.retry = retry
        self.request_timeout = request_timeout
        self.metadata = metadata
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def poke(self, context: Context):
        hook = WorkflowsHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)
        self.log.info("Checking state of execution %s for workflow %s", self.execution_id, self.workflow_id)
        execution: Execution = hook.get_execution(
            workflow_id=self.workflow_id,
            execution_id=self.execution_id,
            location=self.location,
            project_id=self.project_id,
            retry=self.retry,
            timeout=self.request_timeout,
            metadata=self.metadata,
        )

        state = execution.state
        if state in self.failure_states:
            message = (
                f"Execution {self.execution_id} for workflow {self.execution_id} "
                f"failed and is in `{state}` state"
            )
            raise AirflowException(message)

        if state in self.success_states:
            self.log.info(
                "Execution %s for workflow %s completed with state: %s",
                self.execution_id,
                self.workflow_id,
                state,
            )
            return True

        self.log.info(
            "Execution %s for workflow %s does not completed yet, current state: %s",
            self.execution_id,
            self.workflow_id,
            state,
        )
        return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""This module contains a Vertex AI Feature Store sensor."""

from __future__ import annotations

import time
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.common.compat.sdk import AirflowException, BaseSensorOperator
from airflow.providers.google.cloud.hooks.vertex_ai.feature_store import FeatureStoreHook

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class FeatureViewSyncSensor(BaseSensorOperator):
    """
    Sensor to monitor the state of a Vertex AI Feature View sync operation.

    :param feature_view_sync_name: The name of the feature view sync operation to monitor. (templated)
    :param location: Required. The Cloud region in which to handle the request. (templated)
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud Platform.
    :param wait_timeout: How many seconds to wait for sync to complete.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials.
    """

    template_fields: Sequence[str] = ("location", "feature_view_sync_name")
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        feature_view_sync_name: str,
        location: str,
        gcp_conn_id: str = "google_cloud_default",
        wait_timeout: int | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.feature_view_sync_name = feature_view_sync_name
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.wait_timeout = wait_timeout
        self.impersonation_chain = impersonation_chain
        self.start_sensor_time: float | None = None

    def execute(self, context: Context) -> None:
        self.start_sensor_time = time.monotonic()
        super().execute(context)

    def _duration(self):
        return time.monotonic() - self.start_sensor_time

    def poke(self, context: Context) -> bool:
        hook = FeatureStoreHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        try:
            response = hook.get_feature_view_sync(
                location=self.location,
                feature_view_sync_name=self.feature_view_sync_name,
            )

            # Check if the sync has completed by verifying end_time exists
            if response.get("end_time", 0) > 0:
                self.log.info(
                    "Feature View sync %s completed. Rows synced: %d, Total slots: %d",
                    self.feature_view_sync_name,
                    int(response.get("sync_summary", "").get("row_synced", "")),
                    int(response.get("sync_summary", "").get("total_slot", "")),
                )
                return True

            if self.wait_timeout and self._duration() > self.wait_timeout:
                raise AirflowException(
                    f"Timeout: Feature View sync {self.feature_view_sync_name} "
                    f"not completed after {self.wait_timeout}s"
                )

            self.log.info("Waiting for Feature View sync %s to complete.", self.feature_view_sync_name)
            return False

        except Exception as e:
            if self.wait_timeout and self._duration() > self.wait_timeout:
                raise AirflowException(
                    f"Timeout: Feature View sync {self.feature_view_sync_name} "
                    f"not completed after {self.wait_timeout}s"
                )
            self.log.info("Error checking sync status, will retry: %s", str(e))
            return False



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains various unit tests for GCP Cloud Logging Sink Operators
"""

from __future__ import annotations

import re
from datetime import datetime
from unittest import mock

import pytest
from google.api_core.exceptions import AlreadyExists, GoogleAPICallError, InvalidArgument, NotFound
from google.cloud.exceptions import GoogleCloudError
from google.cloud.logging_v2.types import LogSink
from google.protobuf.field_mask_pb2 import FieldMask

from airflow import DAG
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.operators.cloud_logging_sink import (
    CloudLoggingCreateSinkOperator,
    CloudLoggingDeleteSinkOperator,
    CloudLoggingListSinksOperator,
    CloudLoggingUpdateSinkOperator,
)

CLOUD_LOGGING_HOOK_PATH = "airflow.providers.google.cloud.operators.cloud_logging_sink.CloudLoggingHook"
TASK_ID = "test-task"
SINK_NAME = "test-sink"
PROJECT_ID = "test-project"
UNIQUE_WRITER_IDENTITY = True

create_test_cases = [
    (
        {
            "name": SINK_NAME,
            "description": "Creating sink with pubsub",
            "destination": "pubsub.googleapis.com/projects/test-project/topics/test-topic",
            "filter": "severity=INFO",
            "exclusions": [
                {
                    "name": "exclude-debug",
                    "description": "Skip debug logs",
                    "filter": "severity=DEBUG",
                    "disabled": True,
                },
                {
                    "name": "exclude-cloudsql",
                    "description": "Skip CloudSQL logs",
                    "filter": 'resource.type="cloudsql_database"',
                    "disabled": False,
                },
            ],
        }
    ),
    (
        {
            "name": SINK_NAME,
            "description": "Creating bq destination",
            "destination": "bigquery.googleapis.com/projects/test-project/datasets/your_dataset",
            "filter": "severity=ERROR",
            "exclusions": [
                {
                    "name": "exclude-healthchecks",
                    "description": "Exclude App Engine health check logs",
                    "filter": 'resource.type="gae_app" AND protoPayload.status.code=200 AND protoPayload.resource="/_ah/health"',
                    "disabled": False,
                },
                {
                    "name": "exclude-load-balancer-logs",
                    "description": "Exclude HTTP 200 logs from load balancer",
                    "filter": 'resource.type="http_load_balancer" AND httpRequest.status=200',
                    "disabled": False,
                },
                {
                    "name": "exclude-gke-events",
                    "description": "Exclude normal Kubernetes events",
                    "filter": 'resource.type="k8s_event" AND jsonPayload.reason="Scheduled"',
                    "disabled": False,
                },
            ],
            "bigquery_options": {"use_partitioned_tables": True},
        }
    ),
]
create_test_ids = ["create_pubsub", "create_bq"]

update_test_cases = [
    (
        {
            "name": "sink-1",
            "destination": "storage.googleapis.com/my-bucket-1",
            "filter": "severity>=ERROR",
            "description": "Storage sink updated",
            "disabled": False,
        },
        {"paths": ["filter", "description", "disabled"]},
    ),
    (
        {
            "name": "sink-2",
            "destination": "pubsub.googleapis.com/projects/my-project/topics/my-topic",
            "filter": 'resource.type="gce_instance"',
            "description": "Pub/Sub sink updated",
            "disabled": True,
        },
        {"paths": ["destination", "disabled"]},
    ),
]

update_test_ids = ["update_storage_sink", "update_pubsub_sink"]


sink = LogSink(name=SINK_NAME, destination="pubsub.googleapis.com/projects/my-project/topics/my-topic")


def _assert_common_template_fields(template_fields):
    assert "project_id" in template_fields
    assert "gcp_conn_id" in template_fields
    assert "impersonation_chain" in template_fields


class TestCloudLoggingCreateSinkOperator:
    def test_template_fields(self):
        operator = CloudLoggingCreateSinkOperator(task_id=TASK_ID, project_id=PROJECT_ID, sink_config=sink)
        assert "sink_config" in operator.template_fields
        assert "unique_writer_identity" in operator.template_fields
        _assert_common_template_fields(operator.template_fields)

    def test_missing_required_params(self):
        with pytest.raises(AirflowException) as excinfo:
            CloudLoggingCreateSinkOperator(
                task_id=TASK_ID,
                sink_config=None,
                project_id=None,
            ).execute(context={})
        assert "Required parameters are missing: ['project_id', 'sink_config']." in str(excinfo.value)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize("sink_config", create_test_cases, ids=create_test_ids)
    def test_create_with_pubsub_sink(self, hook_mock, sink_config):
        hook_instance = hook_mock.return_value
        hook_instance.create_sink.return_value = LogSink(**sink_config)

        operator = CloudLoggingCreateSinkOperator(
            task_id=TASK_ID,
            sink_config=sink_config,
            project_id=PROJECT_ID,
            unique_writer_identity=UNIQUE_WRITER_IDENTITY,
        )

        operator.execute(context=mock.MagicMock())

        hook_instance.create_sink.assert_called_once_with(
            project_id=PROJECT_ID, sink=sink_config, unique_writer_identity=UNIQUE_WRITER_IDENTITY
        )

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize("sink_config", create_test_cases, ids=create_test_ids)
    def test_create_sink_already_exists(self, hook_mock, sink_config):
        hook_instance = hook_mock.return_value
        hook_instance.create_sink.side_effect = AlreadyExists("Sink already exists")
        hook_instance.get_sink.return_value = LogSink(**sink_config)

        operator = CloudLoggingCreateSinkOperator(
            task_id=TASK_ID,
            sink_config=sink_config,
            project_id=PROJECT_ID,
        )

        result = operator.execute(context=mock.MagicMock())

        hook_instance.create_sink.assert_called_once_with(
            project_id=PROJECT_ID, sink=sink_config, unique_writer_identity=False
        )
        assert result["name"] == sink_config["name"]
        assert result["destination"] == sink_config["destination"]

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize("sink_config", create_test_cases, ids=create_test_ids)
    def test_create_sink_raises_error(self, hook_mock, sink_config):
        hook_instance = hook_mock.return_value
        hook_instance.create_sink.side_effect = GoogleCloudError("Failed to create sink")
        hook_instance.get_sink.return_value = sink_config

        operator = CloudLoggingCreateSinkOperator(
            task_id=TASK_ID,
            sink_config=sink_config,
            project_id=PROJECT_ID,
        )

        with pytest.raises(GoogleCloudError, match="Failed to create sink"):
            operator.execute(context=mock.MagicMock())

        hook_instance.create_sink.assert_called_once()

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(
        "impersonation_chain",
        [
            ["user1@project.iam.gserviceaccount.com", "user2@project.iam.gserviceaccount.com"],
            "user2@project.iam.gserviceaccount.com",
        ],
    )
    def test_create_with_impersonation_chain(self, hook_mock, impersonation_chain):
        hook_instance = hook_mock.return_value
        hook_instance.create_sink.return_value = sink

        operator = CloudLoggingCreateSinkOperator(
            task_id=TASK_ID,
            sink_config=sink,
            impersonation_chain=impersonation_chain,
            project_id=PROJECT_ID,
        )

        operator.execute(context=mock.MagicMock())

        hook_mock.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=impersonation_chain,
        )

    def test_missing_rendered_field_raises(self):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(1997, 9, 25),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingCreateSinkOperator(
                task_id=TASK_ID,
                sink_config="{{ var.value.sink_config }}",
                project_id="{{ var.value.project_id }}",
                dag=dag,
            )

        context = {
            "var": {"value": {"project_id": PROJECT_ID, "sink_config": None}},
        }

        operator.render_template_fields(context)

        with pytest.raises(
            AirflowException,
            match=re.escape(
                "Required parameters are missing: ['sink_config']. These must be passed as keyword parameters."
            ),
        ):
            operator.execute(context)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize("sink_config", create_test_cases, ids=create_test_ids)
    def test_template_rendering(self, hook_mock, sink_config):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingCreateSinkOperator(
                task_id=TASK_ID,
                sink_config="{{ var.value.sink_config }}",
                project_id="{{ var.value.project_id }}",
                dag=dag,
            )

        context = {
            "var": {"value": {"project_id": PROJECT_ID, "sink_config": sink_config}},
        }

        hook_instance = hook_mock.return_value
        hook_instance.create_sink.return_value = LogSink(**sink_config)

        operator.render_template_fields(context)

        operator.execute(context)

        assert isinstance(operator.sink_config, dict)
        assert operator.sink_config == sink_config

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_create_with_empty_sink_name_raises(self, hook_mock):
        sink.name = None
        hook_instance = hook_mock.return_value
        hook_instance.create_sink.side_effect = InvalidArgument("Required parameter 'sink.name' is empty")
        with pytest.raises(
            InvalidArgument,
            match="400 Required parameter 'sink.name' is empty",
        ):
            CloudLoggingCreateSinkOperator(task_id=TASK_ID, sink_config=sink, project_id=PROJECT_ID).execute(
                context={}
            )


class TestCloudLoggingDeleteSinkOperator:
    def test_template_fields(self):
        operator = CloudLoggingDeleteSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            project_id=PROJECT_ID,
        )
        assert "sink_name" in operator.template_fields
        _assert_common_template_fields(operator.template_fields)

    def test_missing_required_params(self):
        with pytest.raises(AirflowException) as excinfo:
            CloudLoggingDeleteSinkOperator(
                task_id=TASK_ID,
                sink_name=None,
                project_id=None,
            ).execute(context={})
        assert "Required parameters are missing" in str(excinfo.value)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_delete_sink_success(self, hook_mock):
        hook_instance = hook_mock.return_value
        hook_instance.delete_sink.return_value = None
        operator = CloudLoggingDeleteSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            project_id=PROJECT_ID,
        )

        context = mock.MagicMock()
        operator.execute(context=context)

        hook_instance.delete_sink.assert_called_once()

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_delete_sink_raises_error(self, hook_mock):
        hook_instance = hook_mock.return_value
        hook_instance.delete_sink.side_effect = GoogleCloudError("Internal Error")

        operator = CloudLoggingDeleteSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            project_id=PROJECT_ID,
        )

        with pytest.raises(GoogleCloudError):
            operator.execute(context=mock.MagicMock())

        hook_instance.delete_sink.assert_called_once()

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_missing_rendered_field_raises(self, hook_mock):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingDeleteSinkOperator(
                task_id=TASK_ID,
                sink_name="{{ var.value.sink_name }}",
                project_id="{{ var.value.project_id }}",
                dag=dag,
            )

        context = {
            "var": {"value": {"project_id": PROJECT_ID, "sink_name": None}},
        }

        operator.render_template_fields(context)
        with pytest.raises(
            AirflowException,
            match=re.escape(
                "Required parameters are missing: ['sink_name']. These must be passed as keyword parameters."
            ),
        ):
            operator.execute(context)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize("sink_config", create_test_cases, ids=create_test_ids)
    def test_template_rendering(self, hook_mock, sink_config):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingDeleteSinkOperator(
                task_id=TASK_ID,
                sink_name="{{ var.value.sink_name }}",
                project_id="{{ var.value.project_id }}",
                dag=dag,
            )

        context = {
            "var": {"value": {"project_id": PROJECT_ID, "sink_name": SINK_NAME}},
        }

        hook_instance = hook_mock.return_value
        hook_instance.delete_sink.return_value = None

        operator.render_template_fields(context)

        operator.execute(context)

        assert operator.project_id == PROJECT_ID
        assert operator.sink_name == SINK_NAME


class TestCloudLoggingListSinksOperator:
    def test_template_fields(self):
        operator = CloudLoggingListSinksOperator(
            task_id=TASK_ID,
            project_id=PROJECT_ID,
        )
        assert "project_id" in operator.template_fields

    def test_missing_required_params(self):
        with pytest.raises(AirflowException) as excinfo:
            CloudLoggingListSinksOperator(
                task_id=TASK_ID,
                project_id=None,
            ).execute(context={})
        assert "Required parameters are missing" in str(excinfo.value)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_list_sinks_success(self, hook_mock):
        hook_instance = hook_mock.return_value
        hook_instance.list_sinks.return_value = [sink, sink]
        operator = CloudLoggingListSinksOperator(
            task_id=TASK_ID,
            project_id=PROJECT_ID,
            page_size=50,
        )
        result = operator.execute(context=mock.MagicMock())
        hook_mock.assert_called_once()
        _, kwargs = hook_mock.call_args
        assert kwargs == {
            "gcp_conn_id": "google_cloud_default",
            "impersonation_chain": None,
        }

        hook_instance.list_sinks.assert_called_once_with(
            project_id=PROJECT_ID,
            page_size=50,
        )

        assert result == [LogSink.to_dict(sink) for sink in [sink, sink]]

    def test_negative_page_size_raises_exception(self):
        with pytest.raises(
            AirflowException, match="The page_size for the list sinks request must be greater than zero"
        ):
            CloudLoggingListSinksOperator(task_id="fail-task", project_id=PROJECT_ID, page_size=-1).execute(
                context={}
            )

    def test_missing_rendered_field_raises(self):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(1997, 9, 25),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingListSinksOperator(
                task_id=TASK_ID, project_id="{{ var.value.project_id }}", dag=dag
            )

        context = {
            "var": {"value": {"project_id": None}},
        }

        operator.render_template_fields(context)

        with pytest.raises(
            AirflowException,
            match=re.escape(
                "Required parameters are missing: ['project_id']. These must be passed as keyword parameters."
            ),
        ):
            operator.execute(context)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    def test_template_rendering(self, hook_mock):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingListSinksOperator(
                task_id=TASK_ID, project_id="{{ var.value.project_id }}", dag=dag
            )

        context = {
            "var": {"value": {"project_id": PROJECT_ID}},
        }

        hook_instance = hook_mock.return_value
        hook_instance.list_sinks.return_value = [sink]

        operator.render_template_fields(context)

        operator.execute(context)

        assert isinstance(operator.project_id, str)
        assert operator.project_id == PROJECT_ID


class TestCloudLoggingUpdateSinksOperator:
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_template_fields(self, sink_config, update_mask):
        operator = CloudLoggingUpdateSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            sink_config=sink_config,
            update_mask=update_mask,
            project_id=PROJECT_ID,
        )
        assert "sink_config" in operator.template_fields
        assert "update_mask" in operator.template_fields
        assert "sink_name" in operator.template_fields
        _assert_common_template_fields(operator.template_fields)

    def test_missing_required_params(self):
        with pytest.raises(AirflowException) as excinfo:
            CloudLoggingDeleteSinkOperator(
                task_id=TASK_ID,
                sink_name=None,
                project_id=None,
            ).execute(context={})
        assert "Required parameters are missing" in str(excinfo.value)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_update_sink_success(self, hook_mock, sink_config, update_mask):
        hook_instance = hook_mock.return_value

        hook_instance.get_sink.return_value = sink
        sink_ = LogSink(**sink_config)
        hook_instance.update_sink.return_value = sink_

        operator = CloudLoggingUpdateSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            project_id=PROJECT_ID,
            sink_config=sink_config,
            update_mask=update_mask,
        )

        result = operator.execute(context=mock.MagicMock())

        hook_instance.get_sink.assert_called_once_with(sink_name=SINK_NAME, project_id=PROJECT_ID)
        hook_instance.update_sink.assert_called_once()
        assert result == LogSink.to_dict(sink_)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_update_sink_raises_not_found(self, hook_mock, sink_config, update_mask):
        hook_instance = hook_mock.return_value

        hook_instance.get_sink.side_effect = NotFound("not found")

        operator = CloudLoggingUpdateSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            sink_config=sink_config,
            update_mask=update_mask,
            project_id=PROJECT_ID,
        )

        with pytest.raises(NotFound, match="not found"):
            operator.execute(context=mock.MagicMock())

        hook_instance.get_sink.assert_called_once()
        hook_instance.update_sink.assert_not_called()

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_update_sink_raises_generic_error(self, hook_mock, sink_config, update_mask):
        hook_instance = hook_mock.return_value
        hook_instance.get_sink.side_effect = GoogleAPICallError("something went wrong")

        operator = CloudLoggingUpdateSinkOperator(
            task_id=TASK_ID,
            sink_name=SINK_NAME,
            sink_config=sink_config,
            update_mask=update_mask,
            project_id=PROJECT_ID,
        )

        with pytest.raises(GoogleAPICallError, match="something went wrong"):
            operator.execute(context=mock.MagicMock())
        hook_instance.get_sink.assert_called_once()
        hook_instance.update_sink.assert_not_called()

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(
        "impersonation_chain",
        [
            ["user1@project.iam.gserviceaccount.com", "user2@project.iam.gserviceaccount.com"],
            "user2@project.iam.gserviceaccount.com",
        ],
    )
    def test_create_with_impersonation_chain(self, hook_mock, impersonation_chain):
        hook_instance = hook_mock.return_value
        hook_instance.get_sink.return_value = sink
        hook_instance.update_sink.return_value = sink

        operator = CloudLoggingUpdateSinkOperator(
            task_id=TASK_ID,
            sink_config=update_test_cases[0][0],
            update_mask=update_test_cases[0][1],
            sink_name=SINK_NAME,
            impersonation_chain=impersonation_chain,
            project_id=PROJECT_ID,
        )

        operator.execute(context=mock.MagicMock())

        hook_mock.assert_called_once_with(
            gcp_conn_id="google_cloud_default",
            impersonation_chain=impersonation_chain,
        )

    def test_missing_rendered_field_raises(self):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(1997, 9, 25),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingUpdateSinkOperator(
                task_id=TASK_ID,
                sink_name="{{ var.value.sink_name }}",
                sink_config="{{ var.value.sink_config }}",
                update_mask="{{ var.value.update_mask }}",
                project_id="{{ var.value.project_id }}",
                dag=dag,
            )

        context = {
            "var": {
                "value": {
                    "project_id": PROJECT_ID,
                    "sink_name": None,
                    "sink_config": None,
                    "update_mask": None,
                }
            },
        }

        operator.render_template_fields(context)

        with pytest.raises(
            AirflowException,
            match=re.escape(
                "Required parameters are missing: ['sink_name', 'sink_config', 'update_mask']. These must be passed as keyword parameters."
            ),
        ):
            operator.execute(context)

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_template_rendering(self, hook_mock, sink_config, update_mask):
        with DAG(
            dag_id="test_render_native",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingUpdateSinkOperator(
                task_id=TASK_ID,
                sink_name="{{ var.value.sink_name }}",
                update_mask="{{ var.value.update_mask }}",
                sink_config="{{ var.value.sink_config }}",
                project_id="{{ var.value.project_id }}",
                unique_writer_identity="{{ var.value.unique_writer_identity }}",
                dag=dag,
            )

        context = {
            "var": {
                "value": {
                    "project_id": PROJECT_ID,
                    "sink_config": sink_config,
                    "sink_name": SINK_NAME,
                    "update_mask": update_mask,
                    "unique_writer_identity": UNIQUE_WRITER_IDENTITY,
                }
            }
        }

        hook_instance = hook_mock.return_value
        hook_instance.get_sink.return_value = LogSink(name=SINK_NAME)
        hook_instance.update_sink.return_value = LogSink(**sink_config)

        operator.render_template_fields(context)

        result = operator.execute(context=mock.MagicMock())

        # Assertions
        assert isinstance(operator.sink_config, dict)
        assert isinstance(operator.update_mask, dict)
        assert isinstance(operator.unique_writer_identity, bool)
        assert operator.sink_config["name"] == sink_config["name"]
        assert result["name"] == sink_config["name"]
        assert operator.update_mask == update_mask

        hook_instance.update_sink.assert_called_once_with(
            project_id=PROJECT_ID,
            sink_name=SINK_NAME,
            sink=sink_config,
            update_mask=update_mask,
            unique_writer_identity=UNIQUE_WRITER_IDENTITY,
        )

    @mock.patch(CLOUD_LOGGING_HOOK_PATH)
    @pytest.mark.parametrize(("sink_config", "update_mask"), update_test_cases, ids=update_test_ids)
    def test_template_rendering_with_proto(self, hook_mock, sink_config, update_mask):
        sink_obj = LogSink(**sink_config)
        mask_obj = FieldMask(paths=update_mask["paths"])
        with DAG(
            dag_id="test_render_native_proto",
            start_date=datetime(2024, 1, 1),
            render_template_as_native_obj=True,
        ) as dag:
            operator = CloudLoggingUpdateSinkOperator(
                task_id=TASK_ID,
                sink_name="{{ var.value.sink_name }}",
                update_mask="{{ var.value.update_mask }}",
                sink_config="{{ var.value.sink_config }}",
                project_id="{{ var.value.project_id }}",
                unique_writer_identity="{{ var.value.unique_writer_identity }}",
                dag=dag,
            )

        context = {
            "var": {
                "value": {
                    "project_id": PROJECT_ID,
                    "sink_name": SINK_NAME,
                    "sink_config": sink_obj,
                    "update_mask": mask_obj,
                    "unique_writer_identity": UNIQUE_WRITER_IDENTITY,
                }
            }
        }

        hook_instance = hook_mock.return_value
        hook_instance.get_sink.return_value = LogSink(name=SINK_NAME)
        hook_instance.update_sink.return_value = sink_obj

        operator.render_template_fields(context)

        result = operator.execute(context=mock.MagicMock())

        assert isinstance(operator.sink_config, LogSink)
        assert isinstance(operator.update_mask, FieldMask)
        assert isinstance(operator.unique_writer_identity, bool)
        assert operator.sink_config.name == sink_obj.name
        assert result["name"] == sink_obj.name
        assert operator.update_mask == mask_obj
        assert operator.sink_config == sink_obj

        hook_instance.update_sink.assert_called_once_with(
            project_id=PROJECT_ID,
            sink_name=SINK_NAME,
            sink=sink_obj,
            update_mask=mask_obj,
            unique_writer_identity=UNIQUE_WRITER_IDENTITY,
        )
